# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-FileCopyrightText: 2024 ANSYS, Inc.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from abc import ABC
from typing import TYPE_CHECKING, List, Union, cast

from ansys.scadeone.core.common.exception import ScadeOneException
from ansys.scadeone.core.common.logger import LOGGER
from ansys.scadeone.core.common.storage import SwanString
from ansys.scadeone.core.common.versioning import gen_swan_version

if TYPE_CHECKING:
    from ansys.scadeone.core import swan
    import ansys.scadeone.core.swan.common as common
    from ansys.scadeone.core.model import Model


class UseDirectiveFactory:
    _instance = None

    def __init__(self) -> None:
        from ansys.scadeone.core.model.loader import SwanParser

        self._parser = SwanParser(LOGGER)

    def __new__(cls, *args, **kwargs) -> "UseDirectiveFactory":
        if not cls._instance:
            cls._instance = super(UseDirectiveFactory, cls).__new__(cls)
        return cls._instance

    def create_use_directive(
        self, module: Union["swan.Module", str], alias: str = None
    ) -> "swan.UseDirective":
        from ansys.scadeone.core.swan import Module, PathIdentifier

        if isinstance(module, str):
            module = Module(PathIdentifier.from_string(module))
        module_name = module.name.as_string
        if not module.name.is_valid:
            raise ScadeOneException(f"Invalid module name: {module_name}")
        code_str = "%s \nuse %s" % (gen_swan_version(), module_name)
        if alias:
            code_str += f" as {alias}"
        code_str += ";"
        use_dir_str = SwanString(code_str, "new_use")
        use_dir = self._parser.module_body(use_dir_str)
        return use_dir.use_directives[0]


class DeclarationFactory:
    _instance = None

    def __init__(self) -> None:
        from ansys.scadeone.core.model.loader import SwanParser

        self._parser = SwanParser(LOGGER)

    def __new__(cls, *args, **kwargs) -> "DeclarationFactory":
        if not cls._instance:
            cls._instance = super(DeclarationFactory, cls).__new__(cls)
        return cls._instance

    def create_declaration(self, decl: str) -> "swan.Declaration":
        """Create a declaration from an expression."""
        if decl[-1] != ";":
            decl += ";"
        decl_str = SwanString(decl, "new_decl")
        declaration = self._parser.declaration(decl_str)
        from ansys.scadeone.core.swan import (
            ConstDeclarations,
            GlobalDeclaration,
            GroupDeclarations,
            SensorDeclarations,
            TypeDeclarations,
        )

        if not isinstance(declaration, GlobalDeclaration):
            raise ScadeOneException(f"Invalid declaration: {decl}")
        if isinstance(declaration, ConstDeclarations):
            return declaration.constants[0]
        elif isinstance(declaration, TypeDeclarations):
            return declaration.types[0]
        elif isinstance(declaration, SensorDeclarations):
            return declaration.sensors[0]
        elif isinstance(declaration, GroupDeclarations):
            return declaration.groups[0]
        raise ScadeOneException(f"Declaration not supported: {decl}")

    @staticmethod
    def create_operator(
        name: Union[List["swan.Identifier"], str],
        is_node: bool = True,
        is_inlined: bool = False,
    ) -> "swan.Operator":
        """Create an operator."""
        from ansys.scadeone.core.swan import Identifier, Operator, Scope

        id = Identifier(name)  # noqa
        if not id.is_valid:
            raise ScadeOneException(f"Invalid module name: {name}")
        return Operator(id, is_inlined, is_node, [], [], Scope())

    @staticmethod
    def create_signature(
        name: Union[List["swan.Identifier"], str],
        is_node: bool = True,
        is_inlined: bool = False,
    ) -> "swan.Signature":
        """Create a signature."""
        from ansys.scadeone.core.swan import Identifier, Signature

        id = Identifier(name)  # noqa
        if not id.is_valid:
            raise ScadeOneException(f"Invalid module name: {name}")
        return Signature(id, is_inlined, is_node, [], [])

    def create_textual_operator(self, operator_str: str) -> "swan.Operator":
        """Create a textual operator."""
        from ansys.scadeone.core.swan import Operator

        operator = self._parser.operator_decl(SwanString(operator_str, "new_op"))
        if not isinstance(operator, Operator):
            raise ScadeOneException(f"Invalid operator: {operator_str}")
        operator.is_text = True
        return operator

    def create_textual_signature(self, operator_str: str) -> "swan.Operator":
        """Create a textual operator signature."""
        from ansys.scadeone.core.swan import Signature

        operator = self._parser.operator_decl(SwanString(operator_str, "new_op"))
        if isinstance(operator, Signature):
            operator.is_text = True
            return operator
        raise ScadeOneException(f"Invalid operator: {operator_str}")


class ModuleFactory:
    @staticmethod
    def create_module(name: Union[List["swan.PathIdentifier"], str]) -> "swan.Module":
        """Create a module."""
        from ansys.scadeone.core.swan import ModuleBody

        id_path = ModuleFactory._get_path_identifier(name)
        return ModuleBody(id_path)

    @staticmethod
    def create_module_interface(
        name: Union[List["swan.PathIdentifier"], str],
    ) -> "swan.ModuleInterface":
        """Create a module interface."""
        from ansys.scadeone.core.swan import ModuleInterface

        id_path = ModuleFactory._get_path_identifier(name)
        return ModuleInterface(id_path)

    @staticmethod
    def _get_path_identifier(name: str) -> "swan.PathIdentifier":
        """Get a path identifier from a module name."""
        from ansys.scadeone.core.swan import Identifier, PathIdentifier

        if isinstance(name, str):
            if name.find("::") == -1:
                path_id = PathIdentifier([Identifier(name)])
            else:
                path = name.split("::")
                path_id = PathIdentifier([Identifier(p) for p in path])
        elif isinstance(name, PathIdentifier):
            path_id = name
        else:
            raise ScadeOneException(f"Invalid module name: {name}")
        if not path_id.is_valid:
            raise ScadeOneException(f"Invalid module name: {name}")
        return path_id


class UseDirectiveAdder:
    @staticmethod
    def add_use_directive(module: "swan.Module", use_directive: "swan.UseDirective") -> None:
        """Add a use directive to the module."""
        module.use_directives.append(use_directive)
        use_directive._owner = module


class DeclarationAdder:
    @staticmethod
    def add_declaration(module: "swan.Module", declaration: "swan.Declaration") -> None:
        """Add a declaration to the module."""
        from ansys.scadeone.core.swan import (
            ConstDecl,
            ConstDeclarations,
            GroupDecl,
            GroupDeclarations,
            ModuleItem,
            Operator,
            SensorDecl,
            SensorDeclarations,
            Signature,
            TypeDecl,
            TypeDeclarations,
        )

        if isinstance(declaration, ConstDecl):
            const_decl = ConstDeclarations([declaration])
            module.declarations.append(cast(ModuleItem, const_decl))
        elif isinstance(declaration, TypeDecl):
            type_decl = TypeDeclarations([declaration])
            module.declarations.append(cast(ModuleItem, type_decl))
        elif isinstance(declaration, SensorDecl):
            sensor_decl = SensorDeclarations([declaration])
            module.declarations.append(cast(ModuleItem, sensor_decl))
        elif isinstance(declaration, GroupDecl):
            group_decl = GroupDeclarations([declaration])
            module.declarations.append(cast(ModuleItem, group_decl))
        elif isinstance(declaration, Operator):
            module.declarations.append(declaration)
        elif isinstance(declaration, Signature):
            module.declarations.append(declaration)
        else:
            raise ScadeOneException(f"Declaration not supported: {declaration}")
        declaration._owner = module


class ModuleAdder:
    @staticmethod
    def add_module(model: "Model", module: "swan.Module") -> None:
        """Add a module to the model."""
        from ansys.scadeone.core.swan import ModuleBody, ModuleInterface, TestHarness

        if model.module_exists(module):
            raise ScadeOneException(f"Module '{module.name}' already exists.")
        if isinstance(module, ModuleBody):
            model.add_body(module)
        elif isinstance(module, ModuleInterface):
            model.add_interface(module)
        elif isinstance(module, TestHarness):
            model.add_harness(module)
        else:
            assert False, f"Module type not supported: {type(module)}"


class ModuleCreator(ABC):
    def use(self, module: Union["swan.Module", str], alias: str = None) -> "swan.UseDirective":
        """Add a use directive to the module.

        Parameters
        ----------
        module: Module
            Module name.
        alias: Str
            Module alias.

        Returns
        -------
        UseDirective
            UseDirective object.
        """
        from ansys.scadeone.core.swan import Module

        if not isinstance(self, Module):
            raise ScadeOneException(f"Cannot add use directive to {type(self)}")
        use_directive = UseDirectiveFactory().create_use_directive(module, alias)
        UseDirectiveAdder.add_use_directive(self, use_directive)
        return use_directive

    def add_declaration(self, decl_expr: str) -> "common.Declaration":
        """Add a declaration to the module.

        Parameters
        ----------
        decl_expr: str
            Declaration expression.

        Returns
        -------
        Declaration
            Declaration object.
        """
        from ansys.scadeone.core.swan import Module

        if not isinstance(self, Module):
            raise ScadeOneException(f"Cannot add sensor to {type(self)}")
        declaration = DeclarationFactory().create_declaration(decl_expr)
        DeclarationAdder.add_declaration(self, declaration)
        return declaration

    def add_constant(self, name: str, type_str: str, value: str) -> "swan.ConstDecl":
        """Add a constant to the module.

        Parameters
        ----------
        name: str
            Constant name.
        type_str: Str
            Constant type.
        value: Str
            Constant value.

        Returns
        -------
        ConstDecl
            Constant object.
        """
        from ansys.scadeone.core.swan import ConstDecl, Module

        if not isinstance(self, Module):
            raise ScadeOneException(f"Cannot add constant to {type(self)}")
        const = DeclarationFactory().create_declaration(f"const {name}: {type_str} = {value}")
        DeclarationAdder.add_declaration(self, const)
        return cast(ConstDecl, const)

    def add_type(self, name: str, type_str: str) -> "swan.TypeDecl":
        """Add a type to the module.

        Parameters
        ----------
        name: str
            Type name.
        type_str: str
            Type definition as a string.

        Returns
        -------
        TypeDecl
            Type object.
        """
        from ansys.scadeone.core.swan import Module, TypeDecl

        if not isinstance(self, Module):
            raise ScadeOneException(f"Cannot add sensor to {type(self)}")
        type_decl = DeclarationFactory().create_declaration(f"type {name} = {type_str}")
        DeclarationAdder.add_declaration(self, type_decl)
        return cast(TypeDecl, type_decl)

    def add_enum(self, name: str, values: List[str]) -> "swan.TypeDecl":
        """Add an enum to the module.

        Parameters
        ----------
        name: str
            Enum name.
        values: List[Str]
            Enum values.

        Returns
        -------
        TypeDecl
            Type object.
        """
        from ansys.scadeone.core.swan import Module, TypeDecl

        if not isinstance(self, Module):
            raise ScadeOneException(f"Cannot add enum to {type(self)}")
        type_decl = DeclarationFactory().create_declaration(
            f"type {name} = enum {{{', '.join(values)}}}"
        )
        DeclarationAdder.add_declaration(self, type_decl)
        return cast(TypeDecl, type_decl)

    def add_struct(self, name: str, fields: dict[str, str]) -> "swan.TypeDecl":
        """Add a struct to the module.

        Parameters
        ----------
        name: str
            Struct name.
        fields: dict[str, str]
            Struct fields (name, type).

        Returns
        -------
        TypeDecl
            Type object.
        """
        from ansys.scadeone.core.swan import Module, TypeDecl

        if not isinstance(self, Module):
            raise ScadeOneException(f"Cannot add struct to {type(self)}")
        decl_str = f"type {name} = {{{', '.join(f'{k}: {v}' for k, v in fields.items())}}}"
        type_decl = DeclarationFactory().create_declaration(decl_str)
        DeclarationAdder.add_declaration(self, type_decl)
        return cast(TypeDecl, type_decl)

    def add_sensor(self, name: str, type_str: str) -> "swan.SensorDecl":
        """Add a sensor to the module.

        Parameters
        ----------
        name: str
            Sensor name.
        type_str: Str
            Sensor type.

        Returns
        -------
        SensorDecl
            Sensor object.
        """
        from ansys.scadeone.core.swan import Module, SensorDecl

        if not isinstance(self, Module):
            raise ScadeOneException(f"Cannot add sensor to {type(self)}")
        sensor = DeclarationFactory().create_declaration(f"sensor {name}: {type_str}")
        DeclarationAdder.add_declaration(self, sensor)
        return cast(SensorDecl, sensor)

    def add_group(
        self, name: str, types: Union[str, List[str], dict[str, str]]
    ) -> "swan.GroupDecl":
        """Add a group to the module.

        Parameters
        ----------
        name: str
            Group name.
        types: Union[str, List[str], dict[str, str]]
            List of types (type or {name: type}).

        Returns
        -------
        GroupDecl
            Group object.
        """
        from ansys.scadeone.core.swan import GroupDecl, Module

        if not isinstance(self, Module):
            raise ScadeOneException(f"Cannot add group to {type(self)}")
        if isinstance(types, str):
            type_list = [types]
        elif isinstance(types, list):
            type_list = types
        elif isinstance(types, dict):
            type_list = [f"{k}: {v}" for k, v in types.items()]
        else:
            raise ScadeOneException(f"Invalid types: {types}")
        group = DeclarationFactory().create_declaration(f"group {name} = ({', '.join(type_list)})")
        DeclarationAdder.add_declaration(self, group)
        return cast(GroupDecl, group)

    def add_operator(
        self,
        name: str,
        is_node: bool = True,
        is_inlined: bool = False,
    ) -> "swan.Operator":
        """Add an operator to the module body.

        Parameters
        ----------
        name: str
            Operator name.
        is_node: Bool
            True if Node, otherwise Function.
        is_inlined: Bool
            True if inline operator.

        Returns
        -------
        Operator
            Operator object.
        """
        from ansys.scadeone.core.swan import ModuleBody

        if not isinstance(self, ModuleBody):
            raise ScadeOneException(f"Cannot add operator to {type(self)}")
        operator = DeclarationFactory().create_operator(name, is_node, is_inlined)
        DeclarationAdder.add_declaration(self, operator)
        return operator

    def add_textual_operator(self, operator_str: str) -> "swan.Operator":
        """Add a textual operator to the module body.

        Parameters
        ----------
        operator_str: str
            Operator string.

        Returns
        -------
        Operator
            Operator object.
        """
        from ansys.scadeone.core.swan import ModuleBody, Operator

        if not isinstance(self, ModuleBody):
            raise ScadeOneException(f"Cannot add operator to {type(self)}")
        operator = DeclarationFactory().create_textual_operator(operator_str)
        DeclarationAdder.add_declaration(self, operator)
        return cast(Operator, operator)

    def add_signature(
        self,
        name: str,
        is_node: bool = True,
        is_inlined: bool = False,
    ) -> "swan.Signature":
        """Add an operator signature to the module.

        Parameters
        ----------
        name: str
            Operator name.
        is_inlined: Bool
            True if inline operator.
        is_node: Bool
            True if Node, otherwise Function.

        Returns
        -------
        Signature
            Signature object.
        """
        from ansys.scadeone.core.swan import Module

        if not isinstance(self, Module):
            raise ScadeOneException(f"Cannot add operator to {type(self)}")
        operator = DeclarationFactory().create_signature(name, is_node, is_inlined)
        DeclarationAdder.add_declaration(self, operator)
        return operator

    def add_textual_signature(self, operator_str: str) -> "swan.Signature":
        """Add a textual operator signature to the module.

        Parameters
        ----------
        operator_str: str
            Operator string.

        Returns
        -------
        Signature
            Signature object.
        """
        from ansys.scadeone.core.swan import Module, Signature

        if not isinstance(self, Module):
            raise ScadeOneException(f"Cannot add operator to {type(self)}")
        operator = DeclarationFactory().create_textual_signature(operator_str)
        DeclarationAdder.add_declaration(self, operator)
        return cast(Signature, operator)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from typing import TYPE_CHECKING, List, Optional

from .common import Pragma, PragmaBase, SwanItem

if TYPE_CHECKING:
    from .namespace import ScopeNamespace  # noqa: F401


class Scope(SwanItem, PragmaBase):  # numpydoc ignore=PR01
    """Scope definition:

    | *data_def* ::= *scope*
    | *scope* ::= { {{*scope_section*}} }"""

    def __init__(
        self,
        sections: Optional[List["ScopeSection"]] = None,
        pragmas: Optional[List[Pragma]] = None,
    ) -> None:
        SwanItem.__init__(self)
        PragmaBase.__init__(self, pragmas)
        self._sections = sections or []
        self._pragmas = pragmas or []
        self.set_owner(self, self._sections)

    @property
    def sections(self) -> List["ScopeSection"]:
        """Scope sections."""
        return self._sections

    def get_declaration(self, name: str):
        """Returns the type, global, operator or variable declaration searching by namespace."""
        from ansys.scadeone.core.swan.namespace import ScopeNamespace

        ns = ScopeNamespace(self)
        return ns.get_declaration(name)


class ScopeSection(SwanItem):  # numpydoc ignore=PR01
    """Base class for scopes."""

    def __init__(self) -> None:
        # Cannot use super() because of multiple inheritance from ProtectedSection
        # which leads to a conflict with the MRO (Method Resolution Order)
        SwanItem.__init__(self)
        self._is_text = False

    @property
    def is_text(self) -> bool:
        """True when section is given from {text%...%text} markup."""
        return self._is_text

    @is_text.setter
    def is_text(self, text_flag: bool):
        self._is_text = text_flag

    def get_declaration(self, name: str):
        from ansys.scadeone.core.swan.namespace import ScopeNamespace

        ns = ScopeNamespace(self)
        return ns.get_declaration(name)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from enum import IntEnum
import logging

from ansys.scadeone.core.common.exception import ScadeOneException


class SDType(IntEnum):
    """
    Defines whether the simulation data file to be made
    is a data source (input) or an oracle
    """

    INPUT = 0
    OUTPUT = 1


class ConverterLogger:
    warnings = 0
    errors = 0
    _logger = None
    SD_VERBOSE = False

    @classmethod
    def getLogger(cls, name: str):
        cls._logger = logging.getLogger(name)
        return cls._logger

    @classmethod
    def debug(cls, txt: str):
        cls._logger.debug(txt)

    @classmethod
    def info(cls, txt: str):
        cls._logger.info(txt)

    @classmethod
    def warning(cls, txt: str):
        cls._logger.warning(txt)
        cls.warnings += 1

    @classmethod
    def error(cls, txt: str):
        cls._logger.error(txt)
        cls.errors += 1

    @classmethod
    def exception(cls, txt: str):
        # cls._logger.exception(txt)
        cls.errors += 1
        raise ScadeOneException(txt)

    @classmethod
    def reset_counts(cls):
        cls.warnings = 0
        cls.errors = 0



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
This module contains the classes for variable declarations:
- VarDecl
- ProtectedVariable, for syntactically incorrect variable definition
"""

from typing import Optional, Union

import ansys.scadeone.core.swan.common as common
from ansys.scadeone.core.swan.expressions import ClockExpr


class VarDecl(common.Declaration, common.Variable):  # numpydoc ignore=PR01
    """Class for variable declaration."""

    def __init__(
        self,
        id: common.Identifier,
        is_clock: Optional[bool] = False,
        is_probe: Optional[bool] = False,
        type: Optional[common.GroupTypeExpression] = None,
        when: Optional[ClockExpr] = None,
        default: Optional[common.Expression] = None,
        last: Optional[common.Expression] = None,
    ) -> None:
        common.Declaration.__init__(self, id)
        self._is_clock = is_clock
        self._is_probe = is_probe
        self._type = type
        self._when = when
        self._default = default
        self._last = last
        self._is_input = False
        self._is_output = False

    @property
    def is_clock(self) -> bool:
        """True when variable is a clock."""
        return self._is_clock

    @property
    def is_probe(self) -> bool:
        """True when variable is a probe."""
        return self._is_probe

    @property
    def type(self) -> Union[common.GroupTypeExpression, None]:
        """Variable type."""
        return self._type

    @property
    def when(self) -> Union[ClockExpr, None]:
        """Variable clock."""
        return self._when

    @property
    def default(self) -> Union[common.Expression, None]:
        """Variable default expression."""
        return self._default

    @property
    def last(self) -> Union[common.Expression, None]:
        """Variable last expression."""
        return self._last

    @property
    def is_input(self) -> bool:
        """True when variable is an input."""
        return self._is_input

    @is_input.setter
    def is_input(self, value: bool) -> None:
        self._is_input = value

    @property
    def is_output(self) -> bool:
        """True when variable is an output."""
        return self._is_output

    @is_output.setter
    def is_output(self, value: bool) -> None:
        self._is_output = value

    @property
    def is_local(self) -> bool:
        """True when variable is local."""
        return not self.is_input and not self.is_output


class ProtectedVariable(common.Variable, common.ProtectedItem):  # numpydoc ignore=PR01
    """Protected variable definition as a string."""

    def __init__(self, data: str) -> None:
        common.ProtectedItem.__init__(self, data, common.Markup.Var)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
This module contains the classes for expressions
"""

from enum import Enum, auto
from typing import List, Optional, Union

from ansys.scadeone.core.common.exception import ScadeOneException
import ansys.scadeone.core.swan.common as common


class UnaryOp(Enum):
    # pylint: disable=invalid-name
    """Unary operators:

    - arithmetic operators
    - logical operators
    - Unit delay operator
    """

    #: (**-**) Unary minus.
    Minus = auto()
    #: (**+**) Unary plus.
    Plus = auto()
    #: (**lnot**) Bitwise not.
    Lnot = auto()
    #: (**not**) Logical not.
    Not = auto()
    #: (**pre**) Unit delay.
    Pre = auto()

    @staticmethod
    def to_str(value: "UnaryOp") -> str:
        if value == UnaryOp.Minus:
            return "-"
        elif value == UnaryOp.Plus:
            return "+"
        elif value == UnaryOp.Lnot:
            return "lnot"
        elif value == UnaryOp.Not:
            return "not"
        elif value == UnaryOp.Pre:
            return "pre"


class BinaryOp(Enum):
    """Binary operators:

    - arithmetic operators
    - relational operators
    - logical operators
    - bitwise operators
    - Initial value, initialed unit delay
    - Concat
    """

    # pylint: disable=invalid-name

    #: (+) Addition.
    Plus = auto()
    #: (-) Subtraction.
    Minus = auto()
    #: (*) Multiplication.
    Mult = auto()
    #: (/) Division.
    Slash = auto()
    #: (**mod**) Modulo.
    Mod = auto()
    #  Relational Expressions
    #: (=) Equal.
    Equal = auto()
    #: (<>) Different.
    Diff = auto()
    #: (<) Less than.
    Lt = auto()
    #: (>) Greater than.
    Gt = auto()
    #: (<=) Less than or equal to.
    Leq = auto()
    #: (>=) Greater than or equal to.
    Geq = auto()
    #  Boolean Expressions
    #: (**and**) Logical and.
    And = auto()
    #: (**or**) Logical or.
    Or = auto()
    #: (**xor**) Logical exclusive or.
    Xor = auto()
    # Bitwise Arithmetic
    #: (**land**) Bitwise and.
    Land = auto()
    #: (**lor**) Bitwise or.
    Lor = auto()
    #: (**lxor**) Bitwise exclusive or.
    Lxor = auto()
    #: (**lsl**) Logical shift left.
    Lsl = auto()
    #: (**lsr**) Logical shift right.
    Lsr = auto()
    # Other Binary
    # (**->**) Initial value.
    Arrow = auto()
    # (**->**) Initial value.
    Pre = auto()
    # (**@**) Array concat.
    Concat = auto()

    @staticmethod
    def to_str(value: "BinaryOp") -> str:
        # pylint disable=too-many-return-statements,no-else-return
        if value == BinaryOp.Plus:
            return "+"
        elif value == BinaryOp.Minus:
            return "-"
        elif value == BinaryOp.Mult:
            return "*"
        elif value == BinaryOp.Slash:
            return "/"
        elif value == BinaryOp.Mod:
            return "mod"
        # Bitwise Arithmetic
        elif value == BinaryOp.Land:
            return "land"
        elif value == BinaryOp.Lor:
            return "lor"
        elif value == BinaryOp.Lxor:
            return "lxor"
        elif value == BinaryOp.Lsl:
            return "lsl"
        elif value == BinaryOp.Lsr:
            return "lsr"
        #  Relational Expressions
        elif value == BinaryOp.Equal:
            return "="
        elif value == BinaryOp.Diff:
            return "<>"
        elif value == BinaryOp.Lt:
            return "<"
        elif value == BinaryOp.Gt:
            return ">"
        elif value == BinaryOp.Leq:
            return "<="
        elif value == BinaryOp.Geq:
            return ">="
        #  Boolean Expressions
        elif value == BinaryOp.And:
            return "and"
        elif value == BinaryOp.Or:
            return "or"
        elif value == BinaryOp.Xor:
            return "xor"
        # Other Binary
        elif value == BinaryOp.Arrow:
            return "->"
        elif value == BinaryOp.Pre:
            return "pre"
        elif value == BinaryOp.Concat:
            return "@"


class PathIdExpr(common.Expression):  # numpydoc ignore=PR01
    """:py:class:`ansys.scadeone.core.swan.PathIdentifier` expression."""

    def __init__(self, path_id: common.PathIdentifier) -> None:
        super().__init__()
        self._path_id = path_id

    @property
    def path_id(self) -> common.PathIdentifier:
        """The identifier expression."""
        return self._path_id


class LastExpr(common.Expression):  # numpydoc ignore=PR01
    """Last expression."""

    def __init__(self, id: common.Identifier) -> None:
        super().__init__()
        self._id = id

    @property
    def id(self) -> common.Identifier:
        """Identifier."""
        return self._id


class LiteralKind(Enum):
    """Literal kind enumeration."""

    # pylint: disable=invalid-name

    #: Boolean literal
    Bool = auto()
    #: Char literal
    Char = auto()
    #: Numeric literal (integer or float, with/without size)
    Numeric = auto()
    #: Erroneous literal
    Error = auto()


class Literal(common.Expression):  # numpydoc ignore=PR01
    """Class for char, numeric, and Boolean expression.

    Boolean value is stored as 'true' or 'false'.

    Char value is a ascii char with its value between simple quotes (ex: 'a')
    or an hexadecimal value.

    Numeric value is INTEGER, TYPED_INTEGER, FLOAT, TYPED_FLOAT
    (see language grammar definition and :py:class:`SwanRE` class).
    """

    def __init__(self, value: str) -> None:
        super().__init__()
        self._value = value
        if common.SwanRE.is_char(self._value):
            self._kind = LiteralKind.Char
        elif common.SwanRE.is_bool(self._value):
            self._kind = LiteralKind.Bool
        elif common.SwanRE.is_numeric(self._value):
            self._kind = LiteralKind.Numeric
        else:
            self._kind = LiteralKind.Error

    @property
    def value(self) -> str:
        """Literal expression."""
        return self._value

    @property
    def is_bool(self) -> bool:
        """Return true when LiteralExpr is a Boolean."""
        return self._kind == LiteralKind.Bool

    @property
    def is_true(self) -> bool:
        """Return true when LiteralExpr is true."""
        return self._kind == LiteralKind.Bool and self._value == "true"

    @property
    def is_char(self) -> bool:
        """Return true when LiteralExpr is a char."""
        return self._kind == LiteralKind.Char

    @property
    def is_numeric(self) -> bool:
        """Return true when LiteralExpr is a numeric."""
        return self._kind == LiteralKind.Numeric

    @property
    def is_integer(self) -> bool:
        """Return true when LiteralExpr is an integer."""
        return self._kind == LiteralKind.Numeric and common.SwanRE.is_integer(self.value)

    @property
    def is_float(self) -> bool:
        """Return true when LiteralExpr is a float."""
        return self._kind == LiteralKind.Numeric and common.SwanRE.is_float(self.value)

    def __str__(self) -> str:
        return self.value


class Pattern(common.SwanItem):  # numpydoc ignore=PR01
    """Base class for patterns."""

    def __init__(self) -> None:
        super().__init__()


class ProtectedPattern(Pattern, common.ProtectedItem):  # numpydoc ignore=PR01
    """Protected pattern expression, i.e., saved as string if
    syntactically incorrect."""

    def __init__(self, data: str) -> None:
        common.ProtectedItem.__init__(self, data)


class ClockExpr(common.SwanItem):  # numpydoc ignore=PR01
    """Clock expressions:

    - Id
    - **not** Id
    - ( Id **match** *pattern*)
    """

    def __init__(
        self,
        id: common.Identifier,
        is_not: Optional[bool] = False,
        pattern: Optional[Pattern] = None,
    ) -> None:
        super().__init__()
        self._id = id
        self._is_not = is_not
        self._pattern = pattern
        if is_not and pattern:
            raise ScadeOneException("ClockExpr: not and pattern together")

    @property
    def id(self) -> common.Identifier:
        """Clock identifier."""
        return self._id

    @property
    def is_not(self) -> bool:
        """**not** id clock expression."""
        return self._is_not

    @property
    def pattern(self) -> Union[Pattern, None]:
        """Matching pattern or None."""
        return self._pattern


class UnaryExpr(common.Expression):  # numpydoc ignore=PR01
    """Expression with unary operators
    :py:class`ansys.scadeone.core.swan.expressions.UnaryOp`."""

    def __init__(self, operator: UnaryOp, expr: common.Expression) -> None:
        super().__init__()
        self._operator = operator
        self._expr = expr

    @property
    def operator(self) -> UnaryOp:
        """Unary operator."""
        return self._operator

    @property
    def expr(self) -> common.Expression:
        """Expression."""
        return self._expr


class BinaryExpr(common.Expression):  # numpydoc ignore=PR01
    """Expression with binary operators
    :py:class`ansys.scadeone.swan.expressions.BinaryOp`."""

    def __init__(
        self,
        operator: BinaryOp,
        left: common.Expression,
        right: common.Expression,
    ) -> None:
        super().__init__()
        self._operator = operator
        self._left = left
        self._right = right

    @property
    def operator(self) -> BinaryOp:
        """Binary operator."""
        return self._operator

    @property
    def left(self) -> common.Expression:
        """Left expression."""
        return self._left

    @property
    def right(self) -> common.Expression:
        """Right expression."""
        return self._right


class WhenClockExpr(common.Expression):  # numpydoc ignore=PR01
    """*expr* **when** *clock_expr* expression"""

    def __init__(self, expr: common.Expression, clock: ClockExpr) -> None:
        super().__init__()
        self._expr = expr
        self._clock = clock

    @property
    def expr(self) -> common.Expression:
        """Expression"""
        return self._expr

    @property
    def clock(self) -> ClockExpr:
        """Clock expression"""
        return self._clock


class WhenMatchExpr(common.Expression):  # numpydoc ignore=PR01
    """*expr* **when match** *path_id* expression"""

    def __init__(self, expr: common.Expression, when: common.PathIdentifier) -> None:
        super().__init__()
        self._expr = expr
        self._when = when

    @property
    def expr(self) -> common.Expression:
        """Expression"""
        return self._expr

    @property
    def when(self) -> ClockExpr:
        """When expression"""
        return self._when


class NumericCast(common.Expression):  # numpydoc ignore=PR01
    """Cast expression: ( *expr* :> *type_expr*)."""

    def __init__(self, expr: common.Expression, type: common.TypeExpression) -> None:
        super().__init__()

        self._expr = expr
        self._type = type

    @property
    def expr(self) -> common.Expression:
        """Expression."""
        return self._expr

    @property
    def type(self) -> common.TypeExpression:
        """Type expression."""
        return self._type


class GroupItem(common.SwanItem):  # numpydoc ignore=PR01
    """Item of a group expression: *group_item* ::= [[ *label* : ]] *expr*."""

    def __init__(self, expr: common.Expression, label: Optional[common.Identifier] = None) -> None:
        super().__init__()
        self._expr = expr
        self._label = label

    @property
    def expr(self) -> common.Expression:
        """Expression."""
        return self._expr

    @property
    def label(self) -> Union[common.Identifier, None]:
        """Group item label."""
        return self._label

    @property
    def has_label(self) -> bool:
        return self._label is not None


class Group(common.SwanItem):  # numpydoc ignore=PR01
    """Group item as a list of GroupItem."""

    def __init__(self, items: List[GroupItem]) -> None:
        super().__init__()
        self._items = items

    @property
    def items(self) -> List[GroupItem]:
        """Group items."""
        return self._items


class GroupConstructor(common.Expression):  # numpydoc ignore=PR01
    """A group expression:
    *group_expr ::= (*group*).
    """

    def __init__(self, group: Group) -> None:
        super().__init__()
        self._group = group

    @property
    def group(self) -> Group:
        return self._group


class GroupRenamingBase(common.SwanItem):
    """Group Renaming Base"""

    pass


class GroupRenaming(GroupRenamingBase):  # numpydoc ignore=PR01
    """Group Renaming: (( Id | Integer)) [: [Id]].

    - Renaming source index as Id or Integer, either a name or a position. For example: *a* or 2.
    - Optional renaming target index:

      - No index
      - Renaming as **:** Id, for example: *a* **:** *b*, 2 **:** *b*
      - Shortcut, example *a* **:** means *a* **:** *a*. Note that *a* **:** *a* is not a shortcut.

    Parameters
    ----------

    source: common.Identifier | LiteralExpr
       Source index.

    renaming: common.Identifier  (optional)
       Renaming as an Identifier.

    is_shortcut: bool (optional)
       Renaming is a shortcut of the form ID.
    """

    def __init__(
        self,
        source: Union[common.Identifier, Literal],
        renaming: Optional[common.Identifier] = None,
        is_shortcut: Optional[bool] = False,
    ) -> None:
        super().__init__()
        self._source = source
        self._renaming = renaming
        self._is_shortcut = is_shortcut

    @property
    def source(self) -> Union[common.Identifier, Literal]:
        """Source selection in group."""
        return self._source

    @property
    def is_shortcut(self) -> bool:
        """True when renaming is a shortcut."""
        return self._is_shortcut

    @property
    def is_valid(self) -> bool:
        """True when renaming is a shortcut with no renaming, or a renaming with no shortcut."""
        if self._renaming and self.is_shortcut:
            # check both id are the same
            return self._source.id == self._renaming.id
        return True

    @property
    def is_by_name(self) -> bool:
        """True when access by name."""
        return isinstance(self._source, common.Identifier)

    @property
    def renaming(self) -> Union[common.Identifier, None]:
        """Renaming in new group. None if no renaming."""
        return self._renaming

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, GroupRenaming):
            return False
        # check sources
        if isinstance(self.source, common.Identifier) ^ isinstance(other.source, common.Identifier):
            # check if both source are the same type
            return False
        if self.source.value != other.source.value:
            return False
        # check renaming
        if isinstance(self.renaming, common.Identifier) ^ isinstance(
            other.renaming, common.Identifier
        ):
            # check if both source are the same type
            return False
        if self.renaming and other.renaming and self.renaming.value != other.renaming.value:
            return False
        if self.is_shortcut != other.is_shortcut:
            return False
        return True

    def __hash__(self) -> int:
        """Hash function for GroupRenaming."""
        return hash((self.source, self.renaming, self.is_shortcut))


class ProtectedGroupRenaming(GroupRenamingBase, common.ProtectedItem):  # numpydoc ignore=PR01
    """Specific class when a renaming is protected for syntax error.

    Source is an adaptation such as: .( {syntax%renaming%syntax} ).
    """

    def __init__(self, data: str, markup: Optional[str] = common.Markup.Syntax) -> None:
        common.ProtectedItem.__init__(self, data, markup)

    @property
    def source(self) -> None:
        """Source selection in group."""
        return None

    @property
    def is_shortcut(self) -> bool:
        """True when renaming is a shortcut."""
        return False

    @property
    def is_valid(self) -> bool:
        """True when renaming is a shortcut with no renaming, or a renaming with no shortcut."""
        return False

    @property
    def is_by_name(self) -> bool:
        """True when access by name."""
        return False

    @property
    def renaming(self) -> Union[common.Identifier, None]:
        """Renaming in new group. None if no renaming."""
        return None


class GroupAdaptation(common.SwanItem):  # numpydoc ignore=PR01
    """Group adaptation: *group_adaptation* ::= . ( *group_renamings* )."""

    def __init__(self, renamings: List[GroupRenaming]) -> None:
        super().__init__()
        self._renamings = renamings

    @property
    def renamings(self) -> List[GroupRenaming]:
        """Renaming list of group adaptation."""
        return self._renamings

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, GroupAdaptation):
            return False
        return self._renamings == other.renamings

    def __hash__(self) -> int:
        """Hash function for GroupAdaptation."""
        return hash(tuple(self._renamings))


class GroupProjection(common.Expression):  # numpydoc ignore=PR01
    """Group projection: *group_expr* ::= *expr* *group_adaptation*."""

    def __init__(self, expr: common.Expression, adaptation: GroupAdaptation) -> None:
        super().__init__()
        self._expr = expr
        self._adaptation = adaptation

    @property
    def expr(self) -> common.Expression:
        """Adapted expression."""
        return self._expr

    @property
    def adaptation(self) -> GroupAdaptation:
        """Expression group adaptation."""
        return self._adaptation


# Composite


class ArrayProjection(common.Expression):  # numpydoc ignore=PR01
    """Static projection: *expr* [*index*], where index is a static expression."""

    def __init__(self, expr: common.Expression, index: common.Expression) -> None:
        super().__init__()
        self._expr = expr
        self._index = index

    @property
    def expr(self) -> common.Expression:
        """Expression."""
        return self._expr

    @property
    def index(self) -> common.Expression:
        """Index expression."""
        return self._index


class StructProjection(common.Expression):  # numpydoc ignore=PR01
    """Static structure field access: *expr* . *label*."""

    def __init__(self, expr: common.Expression, label: common.Identifier) -> None:
        super().__init__()
        self._expr = expr
        self._label = label

    @property
    def expr(self) -> common.Expression:
        """Expression."""
        return self._expr

    @property
    def label(self) -> common.Identifier:
        """Field name."""
        return self._label


class StructDestructor(common.Expression):  # numpydoc ignore=PR01
    """Group creation: *path_id* **group** (*expr*)."""

    def __init__(self, group: common.PathIdentifier, expr: common.Expression) -> None:
        super().__init__()
        self._group = group
        self._expr = expr

    @property
    def expr(self) -> common.Expression:
        """Expression."""
        return self._expr

    @property
    def group(self) -> common.PathIdentifier:
        """Group type."""
        return self._group


class Slice(common.Expression):  # numpydoc ignore=PR01
    """Slice expression: *expr* [ *expr* .. *expr*]."""

    def __init__(
        self, expr: common.Expression, start: common.Expression, end: common.Expression
    ) -> None:
        super().__init__()
        self._expr = expr
        self._start = start
        self._end = end

    @property
    def expr(self) -> common.Expression:
        """Expression."""
        return self._expr

    @property
    def start(self) -> common.Expression:
        """Start of slice expression."""
        return self._start

    @property
    def end(self) -> common.Expression:
        """End of slice expression."""
        return self._end


class LabelOrIndex(common.Expression):  # numpydoc ignore=PR01
    """Stores an index as:

    - a label :py:class:`ansys.scadeone.swan.Identifier` or,
    - an expression :py:class:`ansys.scadeone.swan.Expression`.
    """

    def __init__(self, value: Union[common.Identifier, common.Expression]) -> None:
        super().__init__()
        self._value = value

    @property
    def is_label(self) -> bool:
        return isinstance(self.value, common.Identifier)

    @property
    def value(self) -> Union[common.Identifier, common.Expression]:
        """Return the index (expression or label)."""
        return self._value


class ProjectionWithDefault(common.Expression):  # numpydoc ignore=PR01
    """Dynamic projection: (*expr* . {{ *label_or_index* }}+ **default** *expr*)."""

    def __init__(
        self,
        expr: common.Expression,
        indices: List[LabelOrIndex],
        default: common.Expression,
    ) -> None:
        super().__init__()
        self._expr = expr
        self._indices = indices
        self._default = default

    @property
    def expr(self) -> common.Expression:
        """Expression."""
        return self._expr

    @property
    def default(self) -> common.Expression:
        """Default value."""
        return self._default

    @property
    def indices(self) -> List[LabelOrIndex]:
        """List of indices."""
        return self._indices


class ArrayRepetition(common.Expression):  # numpydoc ignore=PR01
    """Array expression: *expr* ^ *expr*."""

    def __init__(self, expr: common.Expression, size: common.Expression) -> None:
        super().__init__()
        self._expr = expr
        self._size = size

    @property
    def expr(self) -> common.Expression:
        """Expression."""
        return self._expr

    @property
    def size(self) -> common.Expression:
        """Array size."""
        return self._size


class ArrayConstructor(common.Expression):  # numpydoc ignore=PR01
    """Array construction expression: [ *group* ]."""

    def __init__(self, group: Group) -> None:
        super().__init__()
        self._group = group

    @property
    def group(self) -> Group:
        """Group items as a Group."""
        return self._group


class StructConstructor(common.Expression):  # numpydoc ignore=PR01
    """Structure expression, with optional type for cast
    to structure from a group: { *group* } [[ : *path_id*]].

    """

    def __init__(self, group: Group, type: Optional[common.PathIdentifier] = None) -> None:
        super().__init__()
        self._group = group
        self._type = type

    @property
    def group(self) -> Group:
        """Group value"""
        return self._group

    @property
    def type(self) -> Union[common.PathIdentifier, None]:
        """Structure type."""
        return self._type


class VariantValue(common.Expression):  # numpydoc ignore=PR01
    """Variant expression: *path_id* { *group* }."""

    def __init__(self, tag: common.PathIdentifier, group: Group) -> None:
        super().__init__()
        self._tag = tag
        self._group = group

    @property
    def group(self) -> Group:
        """Group value."""
        return self._group

    @property
    def tag(self) -> common.PathIdentifier:
        """Variant tag."""
        return self._tag


class Modifier(common.SwanItem):  # numpydoc ignore=PR01
    """Modifier expression: {{ *label_or_index* }}+ = *expr*.

    Label of index can be syntactically incorrect. In which case, _modifier_ is
    a string, and *is_protected* property is True.

    See :py:class:`FunctionalUpdate`.
    """

    def __init__(self, modifier: Union[List[LabelOrIndex], str], expr: common.Expression) -> None:
        super().__init__()
        self._modifier = modifier
        self._expr = expr
        self._is_protected = isinstance(modifier, str)

    @property
    def expr(self) -> common.Expression:
        """Modifier expression."""
        return self._expr

    @property
    def modifier(self) -> Union[List[LabelOrIndex], str]:
        """Modifier as label or index."""
        return self._modifier

    @property
    def is_protected(self):
        """Modifier has a syntax error and is protected."""
        return self._is_protected


class FunctionalUpdate(common.Expression):  # numpydoc ignore=PR01
    """Copy with modification: ( *expr*  **with** *modifier* {{ ; *modifier* }} [[ ; ]] )."""

    def __init__(self, expr: common.Expression, modifiers: List[Modifier]) -> None:
        super().__init__()
        self._expr = expr
        self._modifiers = modifiers

    @property
    def expr(self) -> common.Expression:
        """Expression."""
        return self._expr

    @property
    def modifiers(self) -> List[Modifier]:
        """Copy modifiers."""
        return self._modifiers


# Switches


class IfteExpr(common.Expression):  # numpydoc ignore=PR01
    """Conditional if/then/else expression: **if** *expr* **then** *expr* **else** *expr*."""

    def __init__(
        self,
        cond_expr: common.Expression,
        then_expr: common.Expression,
        else_expr: common.Expression,
    ) -> None:
        super().__init__()
        self._cond = cond_expr
        self._then = then_expr
        self._else = else_expr

    @property
    def cond_expr(self) -> common.Expression:
        """Condition expression."""
        return self._cond

    @property
    def then_expr(self) -> common.Expression:
        """Then expression."""
        return self._then

    @property
    def else_expr(self) -> common.Expression:
        """Else expression."""
        return self._else


class CaseBranch(common.SwanItem):  # numpydoc ignore=PR01
    """Case branch expression:  *pattern* : *expr*.

    See :py:class:`ansys.scadeone.swan.expressions.CaseExpr`."""

    def __init__(self, pattern: Pattern, expr: common.Expression) -> None:
        super().__init__()
        self._pattern = pattern
        self._expr = expr

    @property
    def pattern(self) -> Pattern:
        """Case branch pattern."""
        return self._pattern

    @property
    def expr(self) -> common.Expression:
        """Case branch expression."""
        return self._expr


class CaseExpr(common.Expression):  # numpydoc ignore=PR01
    """Case expression: **case** *expr* **of** {{ | *pattern* : *expr* }}+ )."""

    def __init__(self, expr: common.Expression, branches: List[CaseBranch]) -> None:
        super().__init__()
        self._expr = expr
        self._branches = branches

    @property
    def expr(self) -> common.Expression:
        """Case expression."""
        return self._expr

    @property
    def branches(self) -> List[CaseBranch]:
        """Case branches."""
        return self._branches


class PathIdPattern(Pattern):  # numpydoc ignore=PR01
    """Simple pattern: *pattern* ::= *path_id*."""

    def __init__(self, path_id: common.PathIdentifier) -> None:
        super().__init__()
        self._path_id = path_id

    @property
    def path_id(self) -> common.PathIdentifier:
        """The path_id of pattern."""
        return self._path_id

    def __str__(self) -> str:
        return str(self.path_id)


class VariantPattern(Pattern):  # numpydoc ignore=PR01
    """Variant pattern:

    *pattern* ::=

    - *path_id* **_**: has_underscore is True
    - *path_id* { } has_underscore is False, has_capture is False
    - *path_id* { Id } : has_capture is True

    """

    def __init__(
        self,
        path_id: common.PathIdentifier,
        captured: Optional[common.Identifier] = None,
        underscore: Optional[bool] = False,
    ) -> None:
        super().__init__()
        self._path_id = path_id
        self._captured = captured
        self._underscore = underscore

    @property
    def path_id(self) -> common.PathIdentifier:
        """The path_id of variant pattern."""
        return self._path_id

    @property
    def underscore(self) -> Union[bool, None]:
        """Underscore as bool or None."""
        return self._underscore

    @property
    def has_underscore(self) -> bool:
        """Variant part is '_'."""
        return self._underscore

    @property
    def has_capture(self) -> bool:
        """The variant pattern has a captured tag."""
        return self._captured is not None

    @property
    def empty_capture(self) -> bool:
        """The variant pattern as an empty {} capture."""
        return not (self.has_underscore or self.has_capture)

    @property
    def captured(self) -> Union[common.Identifier, None]:
        """The variant pattern captured tag."""
        return self._captured

    def __str__(self):
        if self.has_underscore:
            return f"{self.path_id} _"
        if self.has_capture:
            return f"{self.path_id} {{{self.captured}}}"
        return f"{self.path_id} {{}}"


class CharPattern(Pattern):  # numpydoc ignore=PR01
    """Pattern: *pattern* ::= CHAR."""

    def __init__(self, value: str) -> None:
        super().__init__()
        self._value = value

    @property
    def value(self) -> str:
        return self._value

    def __str__(self):
        return self.value


class IntPattern(Pattern):  # numpydoc ignore=PR01
    """Pattern: *pattern* ::= [-] INTEGER | [-] TYPED_INTEGER."""

    def __init__(self, value: str, is_minus: Optional[bool] = False) -> None:
        super().__init__()
        self._value = value
        self._is_minus = is_minus

    @property
    def value(self) -> str:
        """Return value as a string, without sign."""
        return self._value

    @property
    def is_minus(self) -> bool:
        """Return True when has sign minus."""
        return self._is_minus

    @property
    def as_int(self) -> int:
        """Return value as an integer."""
        description = common.SwanRE.parse_integer(self.value, self.is_minus)
        return description.value

    def __str__(self):
        return f"-{self.value}" if self.is_minus else self.value


class BoolPattern(Pattern):  # numpydoc ignore=PR01
    """Pattern: *pattern* ::= **true** | **false**."""

    def __init__(self, value: bool) -> None:
        super().__init__()
        self._value = value

    @property
    def value(self) -> bool:
        """Return value as a bool value"""
        return self._value

    @property
    def is_true(self) -> bool:
        """Return True when pattern is **true**, else False."""
        return self._value

    def __str__(self):
        return "true" if self.is_true else "false"


class UnderscorePattern(Pattern):  # numpydoc ignore=PR01
    """Pattern: *pattern* ::= **_**."""

    def __init__(self) -> None:
        super().__init__()

    def __str__(self) -> str:
        return "_"


class DefaultPattern(Pattern):  # numpydoc ignore=PR01
    """Pattern: *pattern* ::= **default**."""

    def __init__(self) -> None:
        super().__init__()

    def __str__(self) -> str:
        return "default"


class PortExpr(common.Expression):  # numpydoc ignore=PR01
    """Port information."""

    def __init__(
        self,
        lunum: Optional[common.Luid] = None,
        luid: Optional[common.Luid] = None,
        is_self: Optional[bool] = False,
    ) -> None:
        super().__init__()
        self._lunum = lunum
        self._luid = luid
        self._is_self = is_self

    @property
    def lunum(self) -> common.Lunum:
        return self._lunum

    @property
    def luid(self) -> common.Luid:
        return self._luid

    @property
    def is_self(self) -> bool:
        return self._is_self


class Window(common.Expression):  # numpydoc ignore=PR01
    """Temporal window: *expr* ::= **window** <<*expr*>> ( *group* ) ( *group* )."""

    def __init__(self, size: common.Expression, init: Group, params: Group) -> None:
        super().__init__()
        self._size = size
        self._params = params
        self._init = init

    @property
    def size(self) -> common.Expression:
        """Window size."""
        return self._size

    @property
    def params(self) -> Group:
        """Window parameters."""
        return self._params

    @property
    def init(self) -> Group:
        """Window initial values."""
        return self._init


class Merge(common.Expression):  # numpydoc ignore=PR01
    """**merge** ( *group* ) {{ ( *group* ) }}."""

    def __init__(self, params: List[Group]) -> None:
        super().__init__()
        self._params = params

    @property
    def params(self) -> List[Group]:
        return self._params


# =============================================
# Protected Items
# =============================================


class ProtectedExpr(common.Expression, common.ProtectedItem):  # numpydoc ignore=PR01
    """Protected expression, i.e., saved as string if syntactically incorrect."""

    def __init__(self, data: str, markup: Optional[str] = common.Markup.Syntax) -> None:
        common.ProtectedItem.__init__(self, data, markup)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
This module contains classes to manipulate types and types expressions.
"""

from typing import List, Optional, Union

import ansys.scadeone.core.swan.common as common


class TypeDefinition(common.SwanItem):  # numpydoc ignore=PR01
    """Base class for type definition classes."""

    pass


class TypeDecl(common.Declaration):  # numpydoc ignore=PR01
    """Type declaration with its name and optional definition:
    *type_decl* ::= id [[ = *type_def* ]]."""

    def __init__(self, id: common.Identifier, definition: Optional[TypeDefinition] = None) -> None:
        super().__init__(id)
        self._definition = definition

    @property
    def definition(self) -> Union[TypeDefinition, None]:
        return self._definition


# Type definitions
# ----------------


class ExprTypeDefinition(TypeDefinition):  # numpydoc ignore=PR01
    """Type definition as a type expression: *type_def* ::= *type_expr*."""

    def __init__(self, type: common.TypeExpression) -> None:
        super().__init__()
        self._type = type

    @property
    def type(self) -> common.TypeExpression:
        return self._type


class EnumTypeDefinition(TypeDefinition):  # numpydoc ignore=PR01
    """Type definition as an enumeration: *type_def* ::= **enum** { id {{ , id }} }."""

    def __init__(self, tags: List[common.Identifier]) -> None:
        super().__init__()
        self._tags = tags

    @property
    def tags(self) -> List[common.Identifier]:
        return self._tags


class PredefinedType(common.TypeExpression):  # numpydoc ignore=PR01
    """Predefined types."""

    def __init__(self) -> None:
        super().__init__()

    @property
    def is_predefined(self) -> bool:
        "True if type is predefined"
        return True

    @property
    def name(self) -> str:
        """Name of a predefined type from its class."""
        return self.__class__.__name__[:-4].lower()


class BoolType(PredefinedType):  # numpydoc ignore=PR01
    """**bool** type."""


class CharType(PredefinedType):  # numpydoc ignore=PR01
    """**char** type."""


class Int8Type(PredefinedType):  # numpydoc ignore=PR01
    """**int8** type."""


class Int16Type(PredefinedType):  # numpydoc ignore=PR01
    """**int16** type."""


class Int32Type(PredefinedType):  # numpydoc ignore=PR01
    """**int32** type."""


class Int64Type(PredefinedType):  # numpydoc ignore=PR01
    """**int64** type."""


class Uint8Type(PredefinedType):  # numpydoc ignore=PR01
    """**uint8** type."""


class Uint16Type(PredefinedType):  # numpydoc ignore=PR01
    """**uin16** type."""


class Uint32Type(PredefinedType):  # numpydoc ignore=PR01
    """**uint32** type."""


class Uint64Type(PredefinedType):  # numpydoc ignore=PR01
    """**uint64** type."""


class Float32Type(PredefinedType):  # numpydoc ignore=PR01
    """**float32** type."""


class Float64Type(PredefinedType):  # numpydoc ignore=PR01
    """**float64** type."""


class SizedTypeExpression(common.TypeExpression):  # numpydoc ignore=PR01
    """Type with a size expression:

    | *type_expr* ::= **signed** << *expr* >>
    |   | **unsigned** << *expr* >>

    """

    def __init__(self, size: common.Expression, is_signed: bool) -> None:
        super().__init__()
        self._expr = size
        self._is_signed = is_signed

    @property
    def is_signed(self):
        return self._is_signed

    @property
    def size(self):
        return self._expr


class TypeReferenceExpression(common.TypeExpression):  # numpydoc ignore=PR01
    """Type reference to another type: *type_expr* ::= *path_id*."""

    def __init__(self, alias: common.PathIdentifier) -> None:
        super().__init__()
        self._alias = alias

    @property
    def alias(self) -> common.PathIdentifier:
        """Returns aliased type name."""
        return self._alias


class VariableTypeExpression(common.TypeExpression):  # numpydoc ignore=PR01
    """Type variable expression:
    *type_expr* ::= 'Id
    """

    def __init__(self, name: common.Identifier) -> None:
        super().__init__()
        self._name = name

    @property
    def name(self) -> common.Identifier:
        """Name of variable."""
        return self._name


class StructField(common.SwanItem):  # numpydoc ignore=PR01
    """Structure field as: ID **:** *type_expr*."""

    def __init__(self, id: common.Identifier, type: common.TypeExpression) -> None:
        super().__init__()
        self._id = id
        self._type = type

    @property
    def id(self) -> common.Identifier:
        """Field name."""
        return self._id

    @property
    def type(self) -> common.TypeExpression:
        """Field type."""
        return self._type


class StructTypeDefinition(TypeDefinition):  # numpydoc ignore=PR01
    """Type definition as a structure: *type_expr* ::= { *field_decl* {{, *field_decl*}}}."""

    def __init__(self, fields: List[StructField]) -> None:
        super().__init__()
        self._fields = fields

    @property
    def fields(self) -> List[StructField]:
        """List of fields."""
        return self._fields


class ArrayTypeExpression(common.TypeExpression):  # numpydoc ignore=PR01
    """Array type expression: *type_expr* := *type_expr* ^ *expr*."""

    def __init__(self, type: common.TypeExpression, size: common.Expression) -> None:
        super().__init__()
        self._type = type
        self._size = size

    @property
    def size(self) -> common.Expression:
        """Array size."""
        return self._size

    @property
    def type(self) -> common.TypeExpression:
        """Array cell type."""
        return self._type


class VariantComponent(common.SwanItem):  # numpydoc ignore=PR01
    """Variant component: *variant* ::= id *variant_type_expr*."""

    def __init__(self, tag: common.Identifier) -> None:
        super().__init__()
        self._tag = tag

    @property
    def tag(self) -> common.Identifier:
        """Variant tag."""
        return self._tag


class VariantSimple(VariantComponent):  # numpydoc ignore=PR01
    """Simple Variant

    *variant* ::= ID {}"""

    def __init__(self, tag: common.Identifier) -> None:
        super().__init__(tag)


class VariantTypeExpr(VariantComponent):  # numpydoc ignore=PR01
    """Variant type expression:

    *variant* ::= ID { *type_expr* }"""

    def __init__(self, tag: common.Identifier, type: common.TypeExpression) -> None:
        super().__init__(tag)
        self._type = type

    @property
    def type(self) -> common.TypeExpression:
        """Variant type expression."""
        return self._type


class VariantStruct(VariantComponent):  # numpydoc ignore=PR01
    """Variant structure expression:

    *variant* ::= ID *struct_texpr*"""

    def __init__(self, tag: common.Identifier, fields: list[StructField]) -> None:
        super().__init__(tag)
        self._fields = fields

    @property
    def fields(self) -> list[StructField]:
        """Variant structure fields."""
        return self._fields


class VariantTypeDefinition(TypeDefinition):  # numpydoc ignore=PR01
    """Type definition as a variant: *type_def* ::= *variant* {{ | *variant* }}."""

    def __init__(self, tags: List[VariantComponent]) -> None:
        super().__init__()
        self._tags = tags

    @property
    def tags(self) -> List[VariantComponent]:
        return self._tags


class ProtectedTypeExpr(common.TypeExpression, common.ProtectedItem):  # numpydoc ignore=PR01
    """Protected type expression, i.e., saved as string if
    syntactically incorrect."""

    def __init__(self, data: str) -> None:
        common.ProtectedItem.__init__(self, data)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
This module contains classes to manipulate group declarations
and group expressions.
"""

from typing import List

import ansys.scadeone.core.swan.common as common


class TypeGroupTypeExpression(common.GroupTypeExpression):  # numpydoc ignore=PR01
    """Group type expression: *group_type_expr* ::= *type_expr*."""

    def __init__(self, type: common.TypeExpression) -> None:
        super().__init__()
        self._type = type

    @property
    def type(self) -> common.TypeExpression:
        """Type expression of group item"""
        return self._type


class NamedGroupTypeExpression(common.GroupTypeExpression):  # numpydoc ignore=PR01
    """A named group type expression, used in GroupTypeExpressionList as id : *group_type_expr*."""

    def __init__(self, label: common.Identifier, type: common.GroupTypeExpression) -> None:
        super().__init__()
        self._label = label
        self._type = type

    @property
    def label(self) -> common.Identifier:
        """Label of named group item."""
        return self._label

    @property
    def type(self) -> common.GroupTypeExpression:
        """Type of named group item."""
        return self._type


class GroupTypeExpressionList(common.GroupTypeExpression):  # numpydoc ignore=PR01
    """Group list made of positional items followed by named items.
    Each item is a group type expression.

    | *group_type_expr* ::= ( *group_type_expr* {{ , *group_type_expr* }}
    |                        {{ , id : *group_type_expr* }} )
    | | ( id : *group_type_expr* {{ , id : *group_type_expr* }} )
    """

    def __init__(
        self,
        positional: List[common.GroupTypeExpression],
        named: List[NamedGroupTypeExpression],
    ) -> None:
        super().__init__()
        self._positional = positional
        self._named = named

    @property
    def positional(self) -> List[common.GroupTypeExpression]:
        """Return positional group items"""
        return [p for p in self._positional]

    @property
    def named(self) -> List[NamedGroupTypeExpression]:
        """Return named group items"""
        return [p for p in self._named]

    @property
    def items(self) -> List[common.GroupTypeExpression]:
        """Returns all items"""
        items = []
        for pos in self.positional:
            items.append(pos)
        for named in self.named:
            items.append(named)
        return items


class GroupDecl(common.Declaration):  # numpydoc ignore=PR01
    """Group declaration with an id and a type."""

    def __init__(self, id: common.Identifier, type: common.GroupTypeExpression) -> None:
        super().__init__(id)
        self._type = type

    @property
    def type(self) -> common.GroupTypeExpression:
        """Group type expression."""
        return self._type
